"""The default mapper, responsible for parsing MARC21 records acording to the
FOLIO community specifications"""
import logging
import sys
import time
import typing
import uuid
from pathlib import Path
from typing import Generator
from typing import List

import i18n
import pymarc
from defusedxml.ElementTree import fromstring
from folio_uuid.folio_namespaces import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient
from pymarc.record import Record

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import IlsFlavour
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.marc_rules_transformation.conditions import Conditions
from folio_migration_tools.marc_rules_transformation.rules_mapper_base import (
    RulesMapperBase,
)


class BibsRulesMapper(RulesMapperBase):
    """Maps a MARC record to inventory instance format according to
    the FOLIO community convention"""

    def __init__(
        self,
        folio_client,
        library_configuration: LibraryConfiguration,
        task_configuration,
    ):
        super().__init__(
            folio_client,
            library_configuration,
            task_configuration,
            self.get_instance_schema(folio_client),
            Conditions(folio_client, self, "bibs", library_configuration.folio_release),
        )
        logging.info("Fetching mapping rules from the tenant")
        rules_endpoint = "/mapping-rules/marc-bib"
        self.mappings = self.folio_client.folio_get_single_object(rules_endpoint)
        logging.info("Fetching valid language codes...")
        self.language_codes = list(self.fetch_language_codes())
        self.instance_relationships: dict = {}
        self.instance_relationship_types: dict = {}
        self.other_mode_of_issuance_id = get_unspecified_mode_of_issuance(self.folio_client)

        self.start = time.time()

    def perform_initial_preparation(self, marc_record: pymarc.Record, legacy_ids):
        folio_instance = {}
        folio_instance["id"] = str(
            FolioUUID(
                str(self.folio_client.okapi_url),
                FOLIONamespaces.instances,
                str(legacy_ids[-1]),
            )
        )
        self.hrid_handler.handle_hrid(
            FOLIONamespaces.instances,
            folio_instance,
            marc_record,
            legacy_ids,
        )
        self.handle_leader_05(marc_record, legacy_ids)
        if self.task_configuration.add_administrative_notes_with_legacy_ids:
            for legacy_id in legacy_ids:
                self.add_legacy_id_to_admin_note(folio_instance, legacy_id)

        return folio_instance

    def handle_leader_05(self, marc_record, legacy_ids):
        leader_05 = marc_record.leader[5] or "Empty"
        self.migration_report.add("RecordStatus", i18n.t("Original value") + f": {leader_05}")
        if leader_05 not in ["a", "c", "d", "n", "p"]:
            marc_record.leader = f"{marc_record.leader[:5]}c{marc_record.leader[6:]}"
            self.migration_report.add(
                "RecordStatus", i18n.t("Changed %{a} to %{b}", a=leader_05, b="c")
            )
        if leader_05 == "d":
            Helper.log_data_issue(legacy_ids, "d in leader. Is this correct?", marc_record.leader)

    def parse_record(
        self, marc_record: pymarc.Record, file_def: FileDefinition, legacy_ids: List[str]
    ) -> list[dict]:
        """Parses a bib recod into a FOLIO Inventory instance object
        Community mapping suggestion: https://bit.ly/2S7Gyp3
         This is the main function

        Args:
            marc_record (pymarc.Record): _description_
            file_def (FileDefinition): _description_
            legacy_ids (List[str]): List of legacy ids in record

        Returns:
            dict: _description_
        """
        self.print_progress()
        ignored_subsequent_fields: set = set()
        bad_tags = set(self.task_configuration.tags_to_delete)  # "907"
        folio_instance = self.perform_initial_preparation(marc_record, legacy_ids)
        for marc_field in marc_record:
            self.report_marc_stats(marc_field, bad_tags, legacy_ids, ignored_subsequent_fields)
            if marc_field.tag not in ignored_subsequent_fields:
                self.process_marc_field(
                    folio_instance,
                    marc_field,
                    ignored_subsequent_fields,
                    legacy_ids,
                )

        self.perform_additional_parsing(folio_instance, marc_record, legacy_ids, file_def)
        clean_folio_instance = self.validate_required_properties(
            "-".join(legacy_ids), folio_instance, self.schema, FOLIONamespaces.instances
        )
        self.dedupe_rec(clean_folio_instance)
        marc_record.remove_fields(*list(bad_tags))
        self.report_folio_mapping(clean_folio_instance, self.schema)
        return [clean_folio_instance]

    def perform_additional_parsing(
        self,
        folio_instance: dict,
        marc_record: Record,
        legacy_ids: List[str],
        file_def: FileDefinition,
    ) -> None:
        """Do stuff not easily captured by the mapping rules

        Args:
            folio_instance (dict): _description_
            marc_record (Record): _description_
            legacy_ids (List[str]): _description_
            file_def (FileDefinition): _description_
        """
        folio_instance["source"] = "MARC"
        folio_instance["instanceFormatIds"] = list(
            set(self.get_instance_format_ids(marc_record, legacy_ids))
        )
        folio_instance["instanceTypeId"] = self.get_instance_type_id(marc_record, legacy_ids)

        folio_instance["modeOfIssuanceId"] = self.get_mode_of_issuance_id(marc_record, legacy_ids)
        self.handle_languages(folio_instance, marc_record, legacy_ids)
        self.handle_suppression(folio_instance, file_def)
        self.handle_holdings(marc_record)
        if prec_titles := folio_instance.get("precedingTitles", []):
            self.migration_report.add("PrecedingSuccedingTitles", f"{len(prec_titles)}")
            del folio_instance["precedingTitles"]
        if succ_titles := folio_instance.get("succeedingTitles", []):
            del folio_instance["succeedingTitles"]
            self.migration_report.add("PrecedingSuccedingTitles", f"{len(succ_titles)}")

    def handle_languages(self, folio_instance, marc_record, legacy_ids):
        if "languages" in folio_instance:
            orig_languages = {lang: None for lang in folio_instance["languages"]}
            orig_languages.update(
                {lang: None for lang in self.get_languages(marc_record, legacy_ids)}
            )
            folio_instance["languages"] = list(orig_languages.keys())
        else:
            folio_instance["languages"] = self.get_languages(marc_record, legacy_ids)
        folio_instance["languages"] = list(
            self.filter_langs(folio_instance["languages"], marc_record, legacy_ids)
        )

    def get_instance_schema(self, folio_client: FolioClient):
        logging.info("Fetching Instance schema...")
        return folio_client.get_from_github(
            "folio-org", "mod-inventory-storage", "ramls/instance.json"
        )

    def handle_holdings(self, marc_record: Record):
        if "852" in marc_record:
            holdingsfields = marc_record.get_fields(
                "852", "866", "867", "868", "865", "864", "863"
            )
            f852s = (f for f in holdingsfields if f.tag == "852")
            f86xs = (
                f for f in holdingsfields if f.tag in ["866", "867", "868", "865", "864", "863"]
            )
            if f852s and not f86xs:
                self.migration_report.add(
                    "HoldingsGenerationFromBibs",
                    i18n.t(
                        "Records with %{has_many}s but no %{has_no}", has_many="852", has_no="86X"
                    ),
                )
            elif any(f852s):
                self.migration_report.add(
                    "HoldingsGenerationFromBibs",
                    i18n.t(
                        "Records with both %{has_many}s and at least one %{has_one}",
                        has_one="86X",
                        has_many="852",
                    ),
                )

            elif any(f86xs):
                self.migration_report.add(
                    "HoldingsGenerationFromBibs",
                    i18n.t("Records without %{has_no}s but with %{has}", has="86X", has_no="852"),
                )

    def wrap_up(self):
        logging.info("Mapper wrapping up")
        if self.task_configuration.update_hrid_settings:
            self.hrid_handler.store_hrid_settings()

    def get_instance_type_id(self, marc_record, legacy_id):
        return_id = ""

        def get_folio_id_by_name(f336a: str):
            match_template = f336a.lower().replace(" ", "")
            match = next(
                (
                    f["id"]
                    for f in self.folio_client.instance_types
                    if f["name"].lower().replace(" ", "") == match_template
                ),
                "",
            )
            if match:
                self.migration_report.add(
                    "RecourceTypeMapping",
                    "336$a - "
                    + i18n.t("Successful matching on %{criteria}", criteria=match_template)
                    + f" ({f336a})",
                )
            else:
                self.migration_report.add(
                    "RecourceTypeMapping",
                    "336$a - "
                    + i18n.t("Unsuccessful matching on %{criteria}", criteria=match_template)
                    + f" ({f336a})",
                )
                Helper.log_data_issue(
                    legacy_id,
                    "instance type name (336$a) -Unsuccessful matching",
                    f336a,
                )
            return match

        if not self.folio_client.instance_types:
            raise TransformationProcessError("", "No instance_types setup in tenant")

        if "336" in marc_record and "b" not in marc_record["336"]:
            self.migration_report.add("RecourceTypeMapping", i18n.t("Subfield b not in 336"))
            if "a" in marc_record["336"]:
                return_id = get_folio_id_by_name(marc_record["336"]["a"])

        if "336" in marc_record and "b" in marc_record["336"]:
            f336_b = marc_record["336"]["b"].lower().replace(" ", "")
            f336_b_norm = f336_b.lower().replace(" ", "")
            t = self.conditions.get_ref_data_tuple_by_code(
                self.folio_client.instance_types,
                "instance_types",
                f336_b_norm,
            )
            if not t:
                self.migration_report.add(
                    "RecourceTypeMapping",
                    "336$b - "
                    + i18n.t(
                        "Code %{code} ('%{code_raw}') not found in FOLIO ",
                        code=f336_b_norm,
                        code_raw=f336_b,
                    ),
                )
                Helper.log_data_issue(
                    legacy_id,
                    i18n.t("instance type code (%{code}) not found in FOLIO", code="336$b"),
                    f336_b,
                )
            else:
                self.migration_report.add(
                    "RecourceTypeMapping",
                    "336$b "
                    + i18n.t(
                        "%{fro} mapped from %{record}", fro=t[1], record=marc_record["336"]["b"]
                    ),
                )
                return_id = t[0]

        if not return_id:
            t = self.conditions.get_ref_data_tuple_by_code(
                self.folio_client.instance_types, "instance_types", "zzz"
            )
            return_id = t[0]
        return return_id

    def get_instance_format_id_by_code(self, legacy_id: str, code: str):
        try:
            match = next(f for f in self.folio_client.instance_formats if f["code"] == code)
            self.migration_report.add(
                "InstanceFormat",
                i18n.t("Successful match") + f"  - {code}->{match['name']}",
            )
            return match["id"]
        except Exception:
            # TODO: Distinguish between generated codes and proper 338bs
            Helper.log_data_issue(legacy_id, "Instance format Code not found in FOLIO", code)
            self.migration_report.add(
                "InstanceFormat",
                i18n.t("Code '%{code}' not found in FOLIO", code=code),
            )
            return ""

    def get_instance_format_id_by_name(self, f337a: str, f338a: str, legacy_id: str):
        f337a = f337a.lower().strip()
        f338a = f338a.lower().strip()
        match_template = f"{f337a} -- {f338a}"
        try:
            match = next(
                f
                for f in self.folio_client.instance_formats
                if f["name"].lower() == match_template
            )
            self.migration_report.add(
                "InstanceFormat",
                i18n.t(
                    "Successful matching on %{criteria_1} and %{criteria_2}",
                    criteria_1="337$a",
                    criteria_2="338$a",
                )
                + f" - {match_template}->{match['name']}",
            )
            return match["id"]
        except Exception:
            Helper.log_data_issue(
                legacy_id,
                "Unsuccessful matching on 337$a and 338$a",
                match_template,
            )
            self.migration_report.add(
                "InstanceFormat",
                i18n.t(
                    "Unsuccessful matching on %{criteria_1} and %{criteria_2}",
                    criteria_1="337$a",
                    criteria_2="338$a",
                )
                + f" - {match_template}",
            )
            return ""

    def f338_source_is_rda_carrier(self, field: pymarc.Field):
        if "2" not in field:
            self.migration_report.add(
                "InstanceFormat",
                ("Instance Format not mapped from field since 338$2 is missing"),
            )
            return False
        elif field["2"].strip().startswith("rdacarrier"):
            return True
        self.migration_report.add(
            "InstanceFormat",
            ("InstanceFormat not mapped since 338$2 (Source) " f"is set to {field['2']}. "),
        )
        return False

    def get_instance_format_ids_from_a(
        self, field_index, f_338: pymarc.Field, all_337s, legacy_id
    ):
        self.migration_report.add(
            "InstanceFormat",
            i18n.t("338$b is missing. Will try parse from 337$a and 338$a"),
        )
        for a in f_338.get_subfields("a"):
            corresponding_337 = all_337s[field_index] if field_index < len(all_337s) else None
            if corresponding_337 and "a" in corresponding_337:
                if fmt_id := self.get_instance_format_id_by_name(
                    corresponding_337["a"], a, legacy_id
                ):
                    yield fmt_id

    def get_instance_format_ids(self, marc_record, legacy_id):
        all_337s = marc_record.get_fields("337")
        all_338s = marc_record.get_fields("338")
        for fidx, f_338 in enumerate(all_338s):
            if self.f338_source_is_rda_carrier(f_338):
                if "b" not in f_338 and "a" in f_338:
                    yield from self.get_instance_format_ids_from_a(
                        fidx, f_338, all_337s, legacy_id
                    )

                for sfidx, b in enumerate(f_338.get_subfields("b")):
                    b = b.replace(" ", "")
                    if len(b) == 2:
                        # Normal 338b. should be able to map this
                        yield self.get_instance_format_id_by_code(legacy_id, b)
                    elif len(b) == 1:
                        corresponding_337 = all_337s[fidx] if fidx < len(all_337s) else None
                        if not corresponding_337:
                            # No matching 337. No use mapping the 338
                            s = i18n.t(
                                "No corresponding 337 to 338 even though 338$b was one character"
                            )
                            Helper.log_data_issue(legacy_id, s, b)
                            self.migration_report.add(
                                "InstanceFormat",
                                s,
                            )
                        else:
                            # Corresponding 337. Try to combine the codes.
                            corresponding_b = (
                                corresponding_337.get_subfields("b")[sfidx]
                                if sfidx < len(corresponding_337.get_subfields("b"))
                                else None
                            )
                            if not corresponding_b:
                                s = i18n.t("No corresponding $b in corresponding 338")
                                Helper.log_data_issue(legacy_id, s, "")
                                self.migration_report.add("InstanceFormat", s)
                            else:
                                combined_code = (corresponding_b + b).strip()
                                if len(combined_code) == 2:
                                    yield self.get_instance_format_id_by_code(
                                        legacy_id, combined_code
                                    )

    def get_mode_of_issuance_id(self, marc_record: Record, legacy_id: List[str]) -> str:
        level = marc_record.leader[7]
        try:
            name = "unspecified"
            if level in ["a", "c", "d", "m"]:
                name = "single unit"
            if level in ["b", "s"]:
                name = "serial"
            if level == "i":
                name = "integrating resource"
            ret = next(
                (
                    i["id"]
                    for i in self.folio_client.modes_of_issuance
                    if name.lower() == i["name"].lower()
                ),
                "",
            )

            self.migration_report.add("MatchedModesOfIssuanceCode", f"{name} -- {ret}")

            if not ret:
                self.migration_report.add(
                    "MatchedModesOfIssuanceCode", i18n.t("Unmatched level") + f": {level}"
                )

                return self.other_mode_of_issuance_id
            return ret
        except IndexError:
            self.migration_report.add(
                "PossibleCleaningTasks", i18n.t("No Leader[7] in") + f" {legacy_id}"
            )

            return self.other_mode_of_issuance_id
        except StopIteration as ee:
            logging.exception(f"{marc_record.leader} {list(self.folio_client.modes_of_issuance)}")
            raise ee from ee

    def get_nature_of_content(self, marc_record: Record) -> List[str]:
        return ["81a3a0e2-b8e5-4a7a-875d-343035b4e4d7"]

    def get_languages_008(self, marc_record: Record):
        if "008" in marc_record and len(marc_record["008"].data) > 38:
            return "".join(marc_record["008"].data[35:38])
        return ""

    def get_languages_041(self, marc_record, legacy_id):
        languages = dict()
        lang_fields = marc_record.get_fields("041")
        if not any(lang_fields):
            return dict()
        subfields = "abdefghjkmn"
        for lang_tag in lang_fields:
            if "2" in lang_tag:
                self.migration_report.add("LanguageCodeSources", lang_tag["2"])
                Helper.log_data_issue(
                    legacy_id, "Field with other Language code", lang_tag.value()
                )
            lang_codes = lang_tag.get_subfields(*list(subfields))
            for lang_code in lang_codes:
                lang_code = str(lang_code).lower().replace(" ", "")
                langlength = len(lang_code)
                if langlength == 3:
                    languages[lang_code.replace(" ", "")] = None
                elif langlength > 3 and langlength % 3 == 0:
                    lc = lang_code.replace(" ", "")
                    new_codes = {lc[i : i + 3]: None for i in range(0, len(lc), 3)}
                    languages.update(new_codes)
        languages = {
            str(lang): None
            for lang in self.filter_langs(
                list(filter(None, languages.keys())), marc_record, legacy_id
            )
            if lang
        }
        return languages

    def get_languages(self, marc_record: Record, legacy_id: str) -> List[str]:
        """Get languages and tranforms them to correct codes

        Args:
            marc_record (Record): _description_
            legacy_id (str): _description_

        Returns:
            List[str]: _description_
        """
        languages = self.get_languages_041(marc_record, legacy_id)
        languages[self.get_languages_008(marc_record)] = None
        for lang in languages.keys():
            self.migration_report.add("LanguagesInRecords", lang)
        return list(languages.keys())

    def fetch_language_codes(self) -> Generator[str, None, None]:
        """Loads the  list of standardized language codes from LoC

        Yields:
            Generator[str, None, None]: _description_
        """
        path = Path(__file__).parent / "loc_language_codes.xml"
        with open(path) as f:
            lines = "".join(f.readlines())
        tree = fromstring(lines)
        name_space = "{info:lc/xmlns/codelist-v1}"
        xpath_expr = "{0}languages/{0}language/{0}code".format(name_space)
        for code in tree.findall(xpath_expr):
            yield code.text

    def filter_langs(
        self, language_values: List[str], marc_record: Record, index_or_legacy_id
    ) -> typing.Generator:
        forbidden_values = ["###", "zxx", "n/a", "N/A", "|||"]
        for language_value in language_values:
            if language_value in self.language_codes and language_value not in forbidden_values:
                yield language_value
            elif language_value == "jap":
                yield "jpn"
            elif language_value == "fra":
                yield "fre"
            elif language_value == "sve":
                yield "swe"
            elif language_value == "tys":
                yield "ger"
            elif not language_value:
                continue
            elif not language_value.strip():
                continue
            elif language_value not in forbidden_values:
                m = "Unrecognized language codes in record"
                Helper.log_data_issue(index_or_legacy_id, m, language_value)
                self.migration_report.add(
                    "UnrecognizedLanguageCodes",
                    f"{m}: {language_value}",
                )

    def get_legacy_ids(self, marc_record: Record, idx: int) -> List[str]:
        ils_flavour: IlsFlavour = self.task_configuration.ils_flavour
        if ils_flavour in {IlsFlavour.sierra, IlsFlavour.millennium}:
            return get_iii_bib_id(marc_record)
        elif ils_flavour == IlsFlavour.tag907y:
            return RulesMapperBase.get_bib_id_from_907y(marc_record, idx)
        elif ils_flavour == IlsFlavour.tagf990a:
            return RulesMapperBase.get_bib_id_from_990a(marc_record, idx)
        elif ils_flavour == IlsFlavour.aleph:
            return self.get_aleph_bib_id(marc_record)
        elif ils_flavour in {IlsFlavour.voyager, "voyager", IlsFlavour.tag001}:
            return RulesMapperBase.get_bib_id_from_001(marc_record, idx)
        elif ils_flavour == IlsFlavour.koha:
            try:
                return [marc_record["999"]["c"]]
            except Exception as e:
                raise TransformationRecordFailedError(
                    idx,
                    "999 $c is missing, although it is required for this legacy ILS choice",
                    marc_record.as_json(),
                ) from e
        elif ils_flavour == IlsFlavour.custom:
            return get_custom_bib_id(marc_record, self.task_configuration.custom_bib_id_field)
        elif ils_flavour == IlsFlavour.none:
            return [str(uuid.uuid4())]
        else:
            raise TransformationProcessError("", f"ILS {ils_flavour} not configured")

    def get_aleph_bib_id(self, marc_record: Record):
        res = {f["b"].strip(): None for f in marc_record.get_fields("998") if "b" in f}
        if any(res):
            self.migration_report.add_general_statistics(
                i18n.t("legacy id from %{fro}", fro="998$b")
            )
            return list(res)
        else:
            try:
                ret = [marc_record["001"].format_field().strip()]
                self.migration_report.add_general_statistics(
                    i18n.t("legacy id from %{fro}", fro="001")
                )
                return ret
            except Exception as e:
                raise TransformationRecordFailedError(
                    "unknown identifier",
                    "001 is missing.although that or 998$b is required for Aleph migrations",
                    marc_record.as_json(),
                ) from e


def get_unspecified_mode_of_issuance(folio_client: FolioClient):
    m_o_is = list(folio_client.modes_of_issuance)
    if not any(m_o_is):
        logging.critical("No Modes of issuance set up in tenant. Quitting...")
        sys.exit(1)
    if not any(i for i in m_o_is if i["name"].lower() == "unspecified"):
        logging.critical(
            "Mode of issuance 'unspecified' missing in tenant "
            "configuration. Please add this to continue. Quitting..."
        )
        sys.exit(1)
    return next(i["id"] for i in m_o_is if i["name"].lower() == "unspecified")


def get_custom_bib_id(marc_record: Record, field_string: str):
    if field_keys := field_string.split("$", maxsplit=1):
        try:
            if len(field_keys) == 2:
                return [marc_record[field_keys[0]][field_keys[1]]]
            else:
                return [marc_record[field_keys[0]]]
        except Exception as e:
            raise TransformationRecordFailedError(
                "unknown identifier",
                f"{field_string} is missing from record but is required in all records",
                marc_record.as_json(),
            ) from e
    else:
        raise TransformationProcessError(
            "", 'Critical process issue. No "customBibIdField" specified in task configuration.'
        )


def get_iii_bib_id(marc_record: Record):
    try:
        return [marc_record["907"]["a"]]
    except Exception as e:
        raise TransformationRecordFailedError(
            "unknown identifier",
            "907 $a is missing, although it is required for Sierra/iii migrations",
            marc_record.as_json(),
        ) from e



import json
import logging
import sys
from os import environ
from pathlib import Path

import httpx
import humps
import i18n
from argparse_prompt import PromptParser
from pydantic import ValidationError

from folio_migration_tools.config_file_load import merge_load
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.migration_tasks import *  # noqa: F403, F401
from folio_migration_tools.migration_tasks import migration_task_base

try:
    i18n.load_config(Path(__file__).parents[2] / "i18n_config.py")
except i18n.I18nFileLoadError:
    i18n.config.settings.update(
        {
            "file_format": "json",
            "skip_locale_root_data": True,
            "fallback": "en",
            "filename_format": "{locale}.{format}",
        }
    )


def parse_args(args):
    task_classes = iter(inheritors(migration_task_base.MigrationTaskBase))
    parser = PromptParser()
    parser.add_argument(
        "configuration_path",
        help="Path to configuration file",
        nargs="?" if "FOLIO_MIGRATION_TOOLS_CONFIGURATION_PATH" in environ else None,
        prompt="FOLIO_MIGRATION_TOOLS_CONFIGURATION_PATH" not in environ,
        default=environ.get("FOLIO_MIGRATION_TOOLS_CONFIGURATION_PATH"),
    )
    tasks_string = ", ".join(sorted(tc.__name__ for tc in task_classes))

    parser.add_argument(
        "task_name",
        help=("Task name. Use one of: " f"{tasks_string}"),
        nargs="?" if "FOLIO_MIGRATION_TOOLS_TASK_NAME" in environ else None,
        prompt="FOLIO_MIGRATION_TOOLS_TASK_NAME" not in environ,
        default=environ.get("FOLIO_MIGRATION_TOOLS_TASK_NAME"),
    )
    parser.add_argument(
        "--okapi_password",
        help="pasword for the tenant in the configuration file",
        prompt="FOLIO_MIGRATION_TOOLS_OKAPI_PASSWORD" not in environ,
        default=environ.get("FOLIO_MIGRATION_TOOLS_OKAPI_PASSWORD"),
        secure=True,
    )
    parser.add_argument(
        "--base_folder_path",
        help=("path to the base folder for this library. Built on migration_repo_template"),
        prompt="FOLIO_MIGRATION_TOOLS_BASE_FOLDER_PATH" not in environ,
        default=environ.get("FOLIO_MIGRATION_TOOLS_BASE_FOLDER_PATH"),
    )
    parser.add_argument(
        "--report_language",
        help=(
            "Language to write the reports. Defaults english for untranslated languages/strings."
        ),
        default=environ.get("FOLIO_MIGRATION_TOOLS_REPORT_LANGUAGE", "en"),
        prompt=False,
    )
    return parser.parse_args(args)


def main():
    try:
        task_classes = list(inheritors(migration_task_base.MigrationTaskBase))

        args = parse_args(sys.argv[1:])
        i18n.set("locale", args.report_language)
        config_file_humped = merge_load(args.configuration_path)
        config_file_humped["libraryInformation"]["okapiPassword"] = args.okapi_password
        config_file_humped["libraryInformation"]["baseFolder"] = args.base_folder_path
        config_file = humps.decamelize(config_file_humped)
        library_config = LibraryConfiguration(**config_file["library_information"])
        try:
            migration_task_config = next(
                t for t in config_file["migration_tasks"] if t["name"] == args.task_name
            )
        except StopIteration:
            task_names = [t.get("name", "") for t in config_file["migration_tasks"]]
            print(
                f"Referenced task name {args.task_name} not found in the "
                f'configuration file. Use one of {", ".join(task_names)}'
                "\nHalting..."
            )
            sys.exit("Task Name Not Found")
        try:
            task_class = next(
                tc
                for tc in task_classes
                if tc.__name__ == migration_task_config["migration_task_type"]
            )
        except StopIteration:
            print(
                f'Referenced task {migration_task_config["migration_task_type"]} '
                "is not a valid option. Update your task to incorporate "
                f"one of {json.dumps([tc.__name__ for tc in task_classes], indent=4)}"
            )
            sys.exit("Task Type Not Found")
        try:
            task_config = task_class.TaskConfiguration(**migration_task_config)
            task_obj = task_class(task_config, library_config)
            task_obj.do_work()
            task_obj.wrap_up()
        except TransformationProcessError as tpe:
            logging.critical(tpe.message)
            print(f"\n{tpe.message}: {tpe.data_value}")
            sys.exit("Transformation Failure")
        logging.info("Work done. Shutting down")
        sys.exit(0)
    except json.decoder.JSONDecodeError as json_error:
        logging.critical(json_error)
        print(json_error.doc)
        print(
            f"\n{json_error}"
            f"\nError parsing the above JSON mapping or configruation file. Halting."
        )
        sys.exit("Invalid JSON")
    except ValidationError as e:
        print(e.json())
        print("Validation errors in configuration file:")
        print("==========================================")

        for validation_message in json.loads(e.json()):
            print(
                f"{validation_message['msg']}\t"
                f"{', '.join(humps.camelize(str(x)) for x in validation_message['loc'])}"
            )
        print("Halting")
        sys.exit("JSON Not Matching Spec")
    except httpx.HTTPError as connection_error:
        print(
            f"\nConnection Error when connecting to {connection_error.request.url}. "
            "Are you connectet to the Internet/VPN? Do you need to update DNS settings?"
        )
        sys.exit("HTTP Not Connecting")
    except FileNotFoundError as fnf_error:
        print(f"\n{fnf_error.strerror}: {fnf_error.filename}")
        sys.exit("File not found")
    except Exception as ee:
        logging.exception("Unhandled exception")
        print(f"\n{ee}")
        sys.exit(ee.__class__.__name__)


def inheritors(base_class):
    subclasses = set()
    work = [base_class]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return subclasses


if __name__ == "__main__":
    main()



import datetime
import logging
import uuid
from zoneinfo import ZoneInfo

from dateutil import tz
from dateutil.parser import parse

from folio_migration_tools.custom_exceptions import TransformationRecordFailedError

utc = ZoneInfo("UTC")


class LegacyRequest(object):
    def __init__(self, legacy_request_dict, tenant_timezone=utc, row=0):
        # validate
        correct_headers = [
            "item_barcode",
            "patron_barcode",
            "request_date",
            "request_expiration_date",
            "comment",
            "request_type",
            "pickup_servicepoint_id",
        ]
        self.errors = []

        for prop in correct_headers:
            if prop not in legacy_request_dict:
                self.errors.append(("Missing properties in legacy data", prop))
            if prop != "comment" and not legacy_request_dict[prop].strip():
                self.errors.append(("Empty properties in legacy data", prop))

        self.item_barcode = legacy_request_dict["item_barcode"].strip()
        self.patron_id = ""
        self.tenant_timezone = tenant_timezone
        self.item_id = ""
        self.instance_id = ""
        self.holdings_record_id = ""
        self.patron_barcode = legacy_request_dict["patron_barcode"].strip()
        self.comment = legacy_request_dict["comment"].strip()
        self.request_type = legacy_request_dict["request_type"].strip()
        self.pickup_servicepoint_id = legacy_request_dict["pickup_servicepoint_id"].strip()
        self.fulfillment_preference = "Hold Shelf"

        if self.request_type not in ["Hold", "Recall", "Page"]:
            self.errors.append((f"{self.request_type} not allowd", "request_type"))

        try:
            temp_request_date: datetime.datetime = parse(legacy_request_dict["request_date"])
            if temp_request_date.tzinfo != tz.UTC:
                temp_request_date = temp_request_date.replace(tzinfo=self.tenant_timezone)
        except Exception:
            self.errors.append(("Parse date failure. Setting UTC NOW", "request_date"))
            temp_request_date = datetime.now(ZoneInfo("UTC"))
        try:
            temp_expiration_date: datetime.datetime = parse(
                legacy_request_dict["request_expiration_date"]
            )
            if temp_expiration_date.tzinfo != tz.UTC:
                temp_expiration_date = temp_expiration_date.replace(tzinfo=self.tenant_timezone)
        except Exception:
            temp_expiration_date = datetime.now(ZoneInfo("UTC"))
            self.errors.append(("Parse date failure. Setting UTC NOW", "request_expiration_date"))
        if temp_expiration_date.hour == 0 and temp_expiration_date.minute == 0:
            temp_expiration_date = temp_expiration_date.replace(hour=23, minute=59)

        self.request_date: datetime.datetime = temp_request_date
        self.request_expiration_date: datetime.datetime = temp_expiration_date
        self.correct_for_1_day_requests()

    def correct_for_1_day_requests(self):
        try:
            if self.request_expiration_date.date() <= self.request_date.date():
                if (
                    self.request_expiration_date.hour == 0
                    and self.request_expiration_date.minute == 0
                ):
                    self.request_expiration_date = self.request_expiration_date.replace(
                        hour=23, minute=59
                    )
                if self.request_date.hour == 0 and self.request_date.minute == 0:
                    self.request_date = self.request_date.replace(hour=0, minute=1)
            self.make_request_utc()
        except Exception as ee:
            logging.error(ee)
            self.errors.append(("Time alignment issues", "both dates"))

    def to_dict(self):
        return {
            "requestLevel": "Item",
            "requestType": self.request_type,
            "fulfilmentPreference": self.fulfillment_preference,
            "requester": {"barcode": self.patron_barcode},
            "requesterId": self.patron_id,
            "item": {"barcode": self.item_barcode},
            "itemId": self.item_id,
            "instanceId": self.instance_id,
            "holdingsRecordId": self.holdings_record_id,
            "requestExpirationDate": self.request_expiration_date.isoformat(),
            "patronComments": self.comment,
            "pickupServicePointId": self.pickup_servicepoint_id,
            "requestDate": self.request_date.isoformat(),
            "id": str(uuid.uuid4()),
        }

    def serialize(self):
        req = self.to_dict()
        required = [
            "instanceId",
            "requesterId",
            "requestType",
            "requestLevel",
            "requestDate",
            "holdingsRecordId",
            "itemId",
            "fulfilmentPreference",
            "pickupServicePointId",
        ]
        if req["requestLevel"] == "Title":
            required = [r for r in required if r not in ["itemId", "holdingsRecordId"]]
        missing = [r for r in required if not req.get(r, "")]
        if any(missing):
            raise TransformationRecordFailedError(
                "", "Required properties missing:" ", ".join(missing)
            )
        return req

    def to_source_dict(self):
        return {
            "item_barcode": self.item_barcode,
            "patron_barcode": self.patron_barcode,
            "request_date": self.request_date.isoformat(),
            "request_expiration_date": self.request_expiration_date.isoformat(),
            "comment": self.comment,
            "request_type": self.request_type,
            "pickup_servicepoint_id": self.pickup_servicepoint_id,
        }

    def make_request_utc(self):
        try:
            if self.tenant_timezone != ZoneInfo("UTC"):
                self.request_date = self.request_date.astimezone(ZoneInfo("UTC"))
                self.request_expiration_date = self.request_expiration_date.astimezone(
                    ZoneInfo("UTC")
                )
        except Exception:
            self.errors.append(("UTC correction issues", "both dates"))



import ast
import copy
import json
import logging
import sys
import uuid
import i18n
from datetime import datetime
from datetime import timezone
from pathlib import Path

from folio_uuid.folio_namespaces import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.extradata_writer import ExtradataWriter
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)
from folio_migration_tools.migration_report import MigrationReport


class MapperBase:
    legacy_id_template = "Identifier(s) from previous system:"
    bib_id_template = "Bib id: "

    def __init__(
        self,
        library_configuration: LibraryConfiguration,
        folio_client: FolioClient,
        parent_id_map: dict[str, tuple] = None,
    ):
        logging.info("MapperBase initiating")
        self.parent_id_map: dict[str, tuple] = parent_id_map
        self.extradata_writer: ExtradataWriter = ExtradataWriter(Path(""))
        self.start_datetime = datetime.now(timezone.utc)
        self.folio_client: FolioClient = folio_client
        self.library_configuration: LibraryConfiguration = library_configuration

        self.mapped_folio_fields: dict = {}
        self.migration_report: MigrationReport = MigrationReport()
        self.num_criticalerrors = 0
        self.num_exeptions = 0
        self.mapped_legacy_fields: dict = {}
        self.schema_properties = None

    def report_legacy_mapping(self, field_name, present, mapped):
        if field_name:
            try:
                self.mapped_legacy_fields[field_name][0] += int(present)
                self.mapped_legacy_fields[field_name][1] += int(mapped)
            except KeyError:
                self.mapped_legacy_fields[field_name] = [int(present), int(mapped)]

    def report_folio_mapping(self, folio_record, schema):
        try:
            for field_name in set(flatten(folio_record)):
                try:
                    self.mapped_folio_fields[field_name][0] += 1
                except KeyError:
                    self.mapped_folio_fields[field_name] = [1]
            if not self.schema_properties:
                self.schema_properties = schema["properties"].keys()

            unmatched_properties = (
                p for p in self.schema_properties if p not in folio_record.keys()
            )
            for prop in unmatched_properties:
                if prop not in self.mapped_folio_fields:
                    self.mapped_folio_fields[prop] = [0]
        except Exception as ee:
            logging.error(ee, stack_info=True)
            raise ee from ee

    def report_legacy_mapping_no_schema(self, legacy_object):
        for field_name, value in legacy_object.items():
            v = 1 if value else 0
            if field_name not in self.mapped_legacy_fields:
                self.mapped_legacy_fields[field_name] = [1, v]
            else:
                self.mapped_legacy_fields[field_name][0] += 1
                self.mapped_legacy_fields[field_name][1] += v

    def report_folio_mapping_no_schema(self, folio_object):
        for field_name in set(flatten(folio_object)):
            if field_name not in self.mapped_folio_fields:
                self.mapped_folio_fields[field_name] = [1, 1]
            else:
                self.mapped_folio_fields[field_name][0] += 1
                self.mapped_folio_fields[field_name][1] += 1

    def get_mapped_name(
        self,
        ref_data_mapping: RefDataMapping,
        legacy_object,
        index_or_id,
        prevent_default=False,
    ):
        try:
            # Get the values in the fields that will be used for mapping
            fieldvalues = [legacy_object.get(k) for k in ref_data_mapping.mapped_legacy_keys]

            # Gets the first line in the map satisfying all legacy mapping values.
            # Case insensitive, strips away whitespace
            right_mapping = ref_data_mapping.get_ref_data_mapping(legacy_object)

            if not right_mapping:
                raise StopIteration()
            self.migration_report.add(
                ref_data_mapping.blurb_id,
                (
                    f'{" - ".join(fieldvalues)} '
                    f'-> {right_mapping[f"folio_{ref_data_mapping.key_type}"]}'
                ),
            )
            return next(v for k, v in right_mapping.items() if k.startswith("folio_"))

        except StopIteration:
            if prevent_default:
                self.migration_report.add(
                    ref_data_mapping.blurb_id,
                    (f"Not to be mapped. " f'(No default) -- {" - ".join(fieldvalues)} -> ""'),
                )
                return ""
            self.migration_report.add(
                ref_data_mapping.blurb_id,
                (
                    f"Unmapped (Default value was set) -- "
                    f'{" - ".join(fieldvalues)} -> {ref_data_mapping.default_name}'
                ),
            )
            return ref_data_mapping.default_name
        except IndexError as exception:
            raise TransformationRecordFailedError(
                index_or_id,
                (
                    f"{ref_data_mapping.name} - folio_{ref_data_mapping.key_type} "
                    f"({ref_data_mapping.mapped_legacy_keys}) {exception} is not "
                    "a recognized field in the legacy data."
                ),
            ) from exception
        except KeyError as exception:
            raise TransformationProcessError(
                index_or_id,
                (
                    f"{ref_data_mapping.name} mapping - folio_{ref_data_mapping.key_type} "
                    f"({ref_data_mapping.mapped_legacy_keys})  is not "
                    f"a recognized field in the legacy data. KeyError: {exception}"
                ),
            ) from exception
        except Exception as exception:
            raise TransformationProcessError(
                index_or_id,
                (
                    f"{ref_data_mapping.name} - folio_{ref_data_mapping.key_type} "
                    f"({ref_data_mapping.mapped_legacy_keys}) {exception}"
                ),
            ) from exception

    def get_mapped_ref_data_value(
        self,
        ref_data_mapping: RefDataMapping,
        legacy_object,
        index_or_id,
        folio_property_name="",
        prevent_default=False,
    ):
        # Gets mapped value from mapping file, translated to the right FOLIO UUID
        try:
            # Get the values in the fields that will be used for mapping
            fieldvalues = [legacy_object.get(k) for k in ref_data_mapping.mapped_legacy_keys]

            # Gets the first line in the map satisfying all legacy mapping values.
            # Case insensitive, strips away whitespace
            right_mapping = ref_data_mapping.get_ref_data_mapping(legacy_object)
            if not right_mapping:
                # Not all fields matched. Could it be a hybrid wildcard map?
                right_mapping = ref_data_mapping.get_hybrid_mapping(legacy_object)

            if not right_mapping:
                raise StopIteration()
            self.migration_report.add(
                ref_data_mapping.blurb_id,
                (
                    f'{" - ".join(fieldvalues)} '
                    f'-> {right_mapping[f"folio_{ref_data_mapping.key_type}"]}'
                ),
            )
            return right_mapping["folio_id"]
        except StopIteration:
            if prevent_default:
                self.migration_report.add(
                    ref_data_mapping.blurb_id,
                    (f"Not to be mapped. " f'(No default) -- {" - ".join(fieldvalues)} -> ""'),
                )
                return ""
            self.migration_report.add(
                ref_data_mapping.blurb_id,
                (
                    f"Unmapped (Default value was set) -- "
                    f'{" - ".join(fieldvalues)} -> {ref_data_mapping.default_name}'
                ),
            )
            return ref_data_mapping.default_id
        except IndexError as exception:
            raise TransformationRecordFailedError(
                index_or_id,
                (
                    f"{ref_data_mapping.name} - folio_{ref_data_mapping.key_type} "
                    f"({ref_data_mapping.mapped_legacy_keys}) {exception} is not "
                    "a recognized field in the legacy data."
                ),
            ) from exception
        except Exception as exception:
            raise TransformationRecordFailedError(
                index_or_id,
                (
                    f"{ref_data_mapping.name} - folio_{ref_data_mapping.key_type} "
                    f"({ref_data_mapping.mapped_legacy_keys}) {exception}"
                ),
            ) from exception

    def handle_transformation_field_mapping_error(self, index_or_id, error):
        self.migration_report.add("FieldMappingErrors", error)
        error.id = error.id or index_or_id
        error.log_it()
        self.migration_report.add_general_statistics(i18n.t("Field Mapping Errors found"))

    def handle_transformation_process_error(self, idx, error: TransformationProcessError):
        self.migration_report.add_general_statistics(i18n.t("Transformation process error"))
        logging.critical("%s\t%s", idx, error)
        print(f"\n{error.message}: {error.data_value}")
        sys.exit(1)

    def handle_transformation_record_failed_error(
        self, records_processed: int, error: TransformationRecordFailedError
    ):
        self.migration_report.add(
            "GeneralStatistics", i18n.t("FAILED Records failed due to an error")
        )
        error.index_or_id = error.index_or_id or records_processed
        error.log_it()
        self.num_criticalerrors += 1
        if (
            self.num_criticalerrors / (records_processed + 1)
            > (self.library_configuration.failed_percentage_threshold / 100)
            and self.num_criticalerrors > self.library_configuration.failed_records_threshold
        ):
            logging.fatal(
                (
                    "Stopping. More than %s critical data errors. "
                    "Threshold reached. Fix error or raise the bar."
                ),
                self.num_criticalerrors,
            )
            logging.error(
                "Errors: %s\terrors/records: %s",
                self.num_criticalerrors,
                (self.num_criticalerrors / (records_processed + 1)),
            )
            sys.exit(1)

    def get_id_map_tuple(self, legacy_id: str, folio_record: dict, object_type: FOLIONamespaces):
        if object_type == FOLIONamespaces.instances:
            return (legacy_id, folio_record["id"], folio_record["hrid"])
        return (legacy_id, folio_record["id"])

    def handle_generic_exception(self, idx, excepion: Exception):
        self.num_exeptions += 1
        print("\n=======ERROR===========")
        print(
            f"Row {idx:,} failed with the following unhandled Exception: {excepion}  "
            f"of type {type(excepion).__name__}"
        )
        logging.error(excepion, exc_info=True)
        if self.num_exeptions > 50:
            logging.fatal(
                "Stopping. More than %s unhandled exceptions. Code needs fixing",
                self.num_exeptions,
            )
            sys.exit(1)

    def setup_boundwith_relationship_map(self, boundwith_relationship_map):
        new_map = {}
        for entry in boundwith_relationship_map:
            if "MFHD_ID" not in entry or not entry.get("MFHD_ID", ""):
                raise TransformationProcessError(
                    "", "Column MFHD_ID missing from Boundwith relationship map", ""
                )
            if "BIB_ID" not in entry or not entry.get("BIB_ID", ""):
                raise TransformationProcessError(
                    "", "Column BIB_ID missing from Boundwith relationship map", ""
                )
            instance_uuid = str(
                FolioUUID(
                    str(self.folio_client.okapi_url),
                    FOLIONamespaces.instances,
                    entry["BIB_ID"],
                )
            )
            mfhd_uuid = str(
                FolioUUID(
                    str(self.folio_client.okapi_url),
                    FOLIONamespaces.holdings,
                    entry["MFHD_ID"],
                )
            )
            new_map[mfhd_uuid] = new_map.get(mfhd_uuid, []) + [instance_uuid]

        return new_map

    def save_id_map_file(self, path, legacy_map: dict):
        with open(path, "w") as legacy_map_file:
            for id_string in legacy_map.values():
                legacy_map_file.write(f"{json.dumps(id_string)}\n")
                self.migration_report.add(
                    "GeneralStatistics", i18n.t("Unique ID:s written to legacy map")
                )
        logging.info("Wrote legacy id map to %s", path)

    @staticmethod
    def validate_required_properties(
        legacy_id, folio_object: dict, schema: dict, object_type: FOLIONamespaces
    ):
        cleaned_folio_object = MapperBase.clean_none_props(folio_object)
        required = []
        missing = []
        if object_type != FOLIONamespaces.note:
            required = schema.get("required", [])
            missing = list(MapperBase.list_missing(required, cleaned_folio_object))
        else:
            required = (
                schema.get("properties", {}).get("notes", {}).get("items", {}).get("required", [])
            )
            for note in cleaned_folio_object.get("notes", []):
                missing.extend(MapperBase.list_missing(required, note))

        if any(missing):
            raise TransformationRecordFailedError(
                legacy_id,
                "One or many required properties empty",
                f"{json.dumps(missing)}",
            )
        cleaned_folio_object.pop("type", None)
        return cleaned_folio_object

    @staticmethod
    def list_missing(required: list, cleaned_folio_object: dict):
        for required_prop in required:
            if required_prop not in cleaned_folio_object:
                yield f"Missing: {required_prop}"
            elif not cleaned_folio_object[required_prop]:
                yield f"Empty: {required_prop}"

    @staticmethod
    def clean_none_props(d: dict):
        clean = {}
        for k, v in d.items():
            if isinstance(v, dict):
                nested = MapperBase.clean_none_props(v)
                if len(nested.keys()) > 0:
                    clean[k] = nested
            elif isinstance(v, list):
                clean[k] = list(filter(None, v))
            elif v is not None:
                clean[k] = v
        return clean

    def add_legacy_id_to_admin_note(self, folio_record: dict, legacy_id: str):
        if not legacy_id:
            raise TransformationFieldMappingError(
                legacy_id, i18n.t("Legacy id is empty"), legacy_id
            )
        if "administrativeNotes" not in folio_record:
            folio_record["administrativeNotes"] = []
        if id_string := next(
            (f for f in folio_record["administrativeNotes"] if MapperBase.legacy_id_template in f),
            None,
        ):
            if legacy_id not in id_string:
                folio_record["administrativeNotes"] = [
                    f
                    for f in folio_record["administrativeNotes"]
                    if MapperBase.legacy_id_template not in f
                ]

                folio_record["administrativeNotes"].append(f"{id_string}, {legacy_id}")
        else:
            folio_record["administrativeNotes"].append(
                f"{MapperBase.legacy_id_template} {legacy_id}"
            )

    def create_and_write_boundwith_part(self, legacy_item_id: str, bound_with_holding_uuid: dict):
        part = {
            "id": str(uuid.uuid4()),
            "holdingsRecordId": bound_with_holding_uuid,
            "itemId": str(
                FolioUUID(
                    self.folio_client.okapi_url,
                    FOLIONamespaces.items,
                    legacy_item_id,
                )
            ),
        }
        self.extradata_writer.write("boundwithPart", part)

    def create_bound_with_holdings(
        self,
        folio_holding: dict,
        instance_ids: list,
        bound_with_holdings_type_id: str,
    ):
        if not bound_with_holdings_type_id:
            raise TransformationProcessError(
                "Missing task setting holdingsTypeUuidForBoundwiths. Add a "
                "holdingstype specifically for boundwith holdings and reference "
                "the UUID in this parameter."
            )
        for bwidx, instance_uuid in enumerate(instance_ids):
            if not instance_uuid:
                raise ValueError(f"No Instance ID for record {folio_holding}")
            bound_with_holding = copy.deepcopy(folio_holding)
            bound_with_holding["instanceId"] = instance_uuid

            if call_number := folio_holding.get("callNumber", None):
                if "[" in call_number:
                    call_numbers = ast.literal_eval(str(folio_holding["callNumber"]))
                    bound_with_holding["callNumber"] = call_numbers[bwidx]
                else:
                    bound_with_holding["callNumber"] = call_number
            bound_with_holding["holdingsTypeId"] = bound_with_holdings_type_id

            # The subsequent copies gets different ids, but the original is maintained.
            if bwidx > 0:
                bound_with_holding["id"] = self.generate_boundwith_holding_uuid(
                    folio_holding["id"], instance_uuid
                )
                if bound_with_holding.get("hrid", ""):
                    bound_with_holding["hrid"] = f'{bound_with_holding["hrid"]}_bw_{bwidx}'
            self.migration_report.add_general_statistics(i18n.t("Bound-with holdings created"))
            yield bound_with_holding

    def generate_boundwith_holding_uuid(self, holding_uuid, instance_uuid):
        return str(
            FolioUUID(
                self.folio_client.okapi_url,
                FOLIONamespaces.holdings,
                f"{holding_uuid}-{instance_uuid}",
            )
        )


def flatten(my_dict: dict, path=""):
    for k, v in iter(my_dict.items()):
        if not path:
            yield k
        if v:
            if isinstance(v, list):
                if path and check_if_list_with_dict_keys(v):
                    yield f"{path}.{k}".strip(".")
                for e in v:
                    if isinstance(e, dict):
                        yield from flatten(dict(e), f"{path}.{k}")
                    elif isinstance(e, str) and path:
                        yield f"{path}.{k}".strip(".")

            elif isinstance(v, dict):
                if path:
                    yield f"{path}.{k}".strip(".")
                yield from flatten(dict(v), f"{path}.{k}")
            elif path:
                yield f"{path}.{k}".strip(".")


def check_if_list_with_dict_keys(data):
    return isinstance(data, list) and all(isinstance(x, dict) for x in data)



import csv
import json
import logging
import sys
import time
import traceback
import i18n
from typing import Dict
from urllib.error import HTTPError

import httpx
from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.custom_dict import InsensitiveDictReader
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.migration_report import MigrationReport
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration
from folio_migration_tools.transaction_migration.legacy_reserve import LegacyReserve


class ReservesMigrator(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        course_reserve_file_path: FileDefinition

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.reserve

    def __init__(
        self,
        task_configuration: TaskConfiguration,
        library_config: LibraryConfiguration,
    ):
        csv.register_dialect("tsv", delimiter="\t")
        self.migration_report = MigrationReport()
        self.valid_reserves = []
        super().__init__(library_config, task_configuration)
        with open(
            self.folder_structure.legacy_records_folder
            / task_configuration.course_reserve_file_path.file_name,
            "r",
            encoding="utf-8",
        ) as reserves_file:
            self.semi_valid_reserves = list(
                self.load_and_validate_legacy_reserves(
                    InsensitiveDictReader(reserves_file, dialect="tsv")
                )
            )
            logging.info(
                "Loaded and validated %s reserves in file",
                len(self.semi_valid_reserves),
            )

            self.valid_reserves = self.semi_valid_reserves
        self.t0 = time.time()
        self.failed: Dict = {}
        logging.info("Init completed")

    def do_work(self):
        logging.info("Starting")
        for num_reserves, legacy_reserve in enumerate(self.valid_reserves, start=1):
            t0_migration = time.time()
            self.migration_report.add_general_statistics(i18n.t("Processed reserves"))
            try:
                self.post_single_reserve(legacy_reserve)
            except Exception as ee:
                logging.exception(
                    f"Error in row {num_reserves}  Reserve: {json.dumps(legacy_reserve)} {ee}"
                )
            if num_reserves % 50 == 0:
                logging.info(f"{timings(self.t0, t0_migration, num_reserves)} {num_reserves}")

    def post_single_reserve(self, legacy_reserve: LegacyReserve):
        try:
            path = f"/coursereserves/courselistings/{legacy_reserve.course_listing_id}/reserves"
            if self.folio_put_post(
                path, legacy_reserve.to_dict(), "POST", i18n.t("Posted reserves")
            ):
                self.migration_report.add_general_statistics(
                    i18n.t("Successfully posted reserves")
                )
            else:
                self.migration_report.add_general_statistics(i18n.t("Failure to post reserve"))
        except Exception as ee:
            logging.error(ee)

    def wrap_up(self):
        self.extradata_writer.flush()
        for k, v in self.failed.items():
            self.failed_and_not_dupe[k] = [v.to_dict()]
        self.migration_report.set("GeneralStatistics", "Failed loans", len(self.failed))
        self.write_failed_reserves_to_file()

        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.migration_report.write_migration_report(
                i18n.t("Reserves migration report"), report_file, self.start_datetime
            )
        self.clean_out_empty_logs()

    def write_failed_reserves_to_file(self):
        # POST /coursereserves/courselistings/40a085bd-b44b-42b3-b92f-61894a75e3ce/reserves
        # Match on legacy course number ()

        csv_columns = ["legacy_identifier", "barcode"]
        with open(self.folder_structure.failed_recs_path, "w+") as failed_reserves_file:
            writer = csv.DictWriter(failed_reserves_file, fieldnames=csv_columns, dialect="tsv")
            writer.writeheader()
            for _k, failed_reserve in self.failed.items():
                writer.writerow(failed_reserve[0])

    def check_barcodes(self):
        """Stub for extension.

        Yields:
            _type_: _description_
        """
        item_barcodes = set()
        self.circulation_helper.load_migrated_item_barcodes(
            item_barcodes, self.task_configuration.item_files, self.folder_structure
        )
        for loan in self.semi_valid_legacy_loans:
            has_item_barcode = loan.item_barcode in item_barcodes or not any(item_barcodes)
            if has_item_barcode:
                self.migration_report.add_general_statistics(
                    i18n.t("Reserve verified against migrated item")
                )
                yield loan
            else:
                self.migration_report.add(
                    "DiscardedLoans", i18n.t("Reserve discarded. Could not find migrated barcode")
                )

    def load_and_validate_legacy_reserves(self, reserves_reader):
        num_bad = 0
        logging.info("Validating legacy loans in file...")
        for legacy_reserve_count, legacy_reserve_dict in enumerate(reserves_reader):
            try:
                legacy_reserve = LegacyReserve(
                    legacy_reserve_dict,
                    self.folio_client,
                    legacy_reserve_count,
                )
                if any(legacy_reserve.errors):
                    num_bad += 1
                    self.migration_report.add_general_statistics(i18n.t("Discarded reserves"))
                    for error in legacy_reserve.errors:
                        self.migration_report.add("DiscardedReserves", f"{error[0]} - {error[1]}")
                else:
                    yield legacy_reserve
            except ValueError as ve:
                logging.exception(ve)
        logging.info(
            f"Done validating {legacy_reserve_count} legacy reserves with {num_bad} rotten apples"
        )
        if num_bad / legacy_reserve_count > 0.5:
            q = num_bad / legacy_reserve_count
            logging.error("%s percent of reserves failed to validate.", (q * 100))
            self.migration_report.log_me()
            logging.critical("Halting...")
            sys.exit(1)

    def folio_put_post(self, url, data_dict, verb, action_description=""):
        full_url = f"{self.folio_client.okapi_url}{url}"
        try:
            if verb == "PUT":
                resp = httpx.put(
                    full_url,
                    headers=self.folio_client.okapi_headers,
                    json=data_dict,
                )
            elif verb == "POST":
                resp = httpx.post(
                    full_url,
                    headers=self.folio_client.okapi_headers,
                    json=data_dict,
                )
            else:
                raise TransformationProcessError("Bad verb supplied. This is a code issue.")
            if resp.status_code == 422:
                error_message = json.loads(resp.text)["errors"][0]["message"]
                logging.error(error_message)
                self.migration_report.add(
                    "Details",
                    i18n.t(
                        "%{action} error: %{message}",
                        action=action_description,
                        message=error_message,
                    ),
                )
                resp.raise_for_status()
            elif resp.status_code in [201, 204]:
                self.migration_report.add(
                    "Details",
                    i18n.t("Successfully %{action}", action=action_description)
                    + f" ({resp.status_code})",
                )
            else:
                self.migration_report.add(
                    "Details",
                    i18n.t(
                        "%{action} error. http status: %{status}",
                        action=action_description,
                        status=resp.status_code,
                    ),
                )
                logging.error(json.dumps(data_dict))
                resp.raise_for_status()
            return True
        except HTTPError as exception:
            logging.error(f"{resp.status_code}. {verb} FAILED for {url}")
            traceback.print_exc()
            logging.info(exception)
            return False


def timings(t0, t0func, num_objects):
    avg = num_objects / (time.time() - t0)
    elapsed = time.time() - t0
    elapsed_func = time.time() - t0func
    return (
        f"Total objects: {num_objects}\tTotal elapsed: {elapsed:.2f}\t"
        f"Average per object: {avg:.2f}\tElapsed this time: {elapsed_func:.2f}"
    )



from humps import camelize
from pydantic import BaseModel


def to_camel(string):
    return camelize(string)


class AbstractTaskConfiguration(BaseModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True



import calendar
import contextlib
import logging
import re
import i18n
from typing import List

from pymarc import Field
from pymarc import Record

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError


class HoldingsStatementsParser:
    @staticmethod
    def get_holdings_statements(
        marc_record: Record,
        pattern_tag: str,
        value_tag: str,
        field_textual: str,
        legacy_ids: List[str],
        dedupe_results: bool = True,
    ) -> dict:
        """_summary_

        Args:
            marc_record (Record): _description_
            pattern_tag (str): _description_
            value_tag (str): _description_
            field_textual (str): _description_
            legacy_ids (List[str]): _description_
            dedupe_results (bool): _description_. Defaults to True.

        Raises:
            TransformationFieldMappingError: _description_

        Returns:
            dict: _description_
        """

        # Textual holdings statements
        return_dict: dict = {"statements": [], "migration_report": [], "hlm_stmts": []}
        HoldingsStatementsParser.get_textual_statements(
            marc_record, field_textual, return_dict, legacy_ids
        )

        value_fields = marc_record.get_fields(value_tag)
        for pattern_field in marc_record.get_fields(pattern_tag):
            if "8" not in pattern_field:
                raise TransformationFieldMappingError(
                    legacy_ids,
                    i18n.t(
                        "%{tag} subfield %{subfield} not in field",
                        tag=pattern_tag,
                        subfield="8",
                    ),
                    pattern_field,
                )
            linked_value_fields = [
                value_field
                for value_field in value_fields
                if "8" in value_field and value_field["8"].split(".")[0] == pattern_field["8"]
            ]

            if not any(linked_value_fields):
                return_dict["migration_report"].append(
                    (
                        "Holdings statements",
                        f"Missing linked fields for {pattern_tag}",
                    )
                )

            else:
                for linked_value_field in linked_value_fields:
                    parsed_dict = HoldingsStatementsParser.parse_linked_field(
                        pattern_field, linked_value_field
                    )
                    if parsed_dict["hlm_stmt"]:
                        return_dict["hlm_stmts"].append(parsed_dict["hlm_stmt"])
                    if parsed_dict["statement"]:
                        logging.info(
                            f"HOLDINGS STATEMENT PATTERN\t{'-'.join(legacy_ids)}\t{pattern_field}"
                            f"\t{linked_value_field}"
                            f"\t{parsed_dict['statement']['statement']}"
                            f"\t{parsed_dict['statement']['note']}"
                            f"\t{parsed_dict['statement']['staffNote']}"
                        )
                        return_dict["migration_report"].append(
                            (
                                "Holdings statements",
                                f"From {pattern_tag}",
                            )
                        )
                        return_dict["statements"].append(parsed_dict["statement"])

        if dedupe_results:
            return_dict["statements"] = HoldingsStatementsParser.dedupe_list_of_dict(
                return_dict["statements"]
            )
        return return_dict

    @staticmethod
    def parse_linked_field(pattern_field: Field, linked_value_fields: Field):
        break_ind = HoldingsStatementsParser.get_break_indicator(linked_value_fields)
        _from, _to, is_span = HoldingsStatementsParser.get_from_to(
            pattern_field, linked_value_fields
        )
        cron_from, cron_to, hlm_stmt, is_cron_span = HoldingsStatementsParser.get_cron_from_to(
            pattern_field, linked_value_fields
        )
        return_dict = {
            "statement": {"statement": "", "note": "", "staffNote": ""},
            "hlm_stmt": hlm_stmt,
        }

        if _from and cron_from:
            _from = f"{_from} ({cron_from})"
        if not _from and cron_from:
            _from = cron_from
        if _to and cron_to:
            _to = f"{_to} ({cron_to})"
        if _to and cron_from and not cron_to:
            _to = f"{_to} ({cron_from})"
        if not _to and cron_to:
            _to = cron_to
        if _to == cron_to and _to:
            _to = f"({cron_to})"
        if _from and _from == cron_from:
            _from = f"({cron_from})"
        span = " - " if is_span or is_cron_span else ""
        stmt = f"{_from}{span}{_to}{break_ind}" if _from else ""
        stmt = stmt.strip()
        if "z" in linked_value_fields:
            return_dict["statement"]["note"] = linked_value_fields["z"]
        if "x" in linked_value_fields:
            return_dict["statement"]["staffNote"] = linked_value_fields["x"]
        stmt = re.sub(" +", " ", stmt)
        return_dict["statement"]["statement"] = stmt
        return return_dict

    @staticmethod
    def get_textual_statements(
        marc_record: Record, field_textual: str, return_dict: dict, legacy_ids: List[str]
    ):
        """Returns the textual statements from the relevant marc fields

        Args:
            marc_record (Record): _description_
            field_textual (str): _description_
            return_dict (dict): _description_
            legacy_ids (List[str]): List of legacy ids in record

        Raises:
            TransformationFieldMappingError: _description_
            TransformationFieldMappingError: _description_
        """
        for f in marc_record.get_fields(field_textual):
            codes = [sf.code for sf in f.subfields]
            if "a" not in codes and "z" not in codes and "x" not in codes:
                raise TransformationFieldMappingError(
                    legacy_ids,
                    i18n.t(
                        "%{field} subfields a, x, and z missing from field", field=field_textual
                    ),
                    f,
                )
            if not (
                len(f.get_subfields("a")) == 0
                or len(f.get_subfields("z")) == 0
                or len(f.get_subfields("x")) == 0
            ):
                raise TransformationFieldMappingError(
                    legacy_ids,
                    i18n.t("%{field} a,x and z are all empty", field=field_textual),
                    f,
                )
            return_dict["statements"].append(
                {
                    "statement": "".join(f.get_subfields("a")),
                    "note": "".join(f.get_subfields("z")),
                    "staffNote": "".join(f.get_subfields("x")),
                }
            )
            return_dict["migration_report"].append(
                ("Holdings statements", f"From {field_textual}")
            )

    @staticmethod
    def get_break_indicator(field: Field):
        if "w" not in field or field["w"] not in ["g", "n"]:
            return ""
        elif field["w"] == "g":
            return ","
        elif field["w"] == "n":
            return ";"
        else:
            return ""

    @staticmethod
    def g_s(val: str):
        val_int = int(val)
        if val_int == 21:
            return "Spring"
        elif val_int == 22:
            return "Summer"
        elif val_int == 23:
            return "Fall"
        elif val_int == 24:
            return "Winter"
        else:
            return val

    @staticmethod
    def get_season(val: str):
        try:
            result = HoldingsStatementsParser.g_s(val)
            if result == val:
                result = HoldingsStatementsParser.g_m(int(val))
            return result
        except Exception:
            return val

    @staticmethod
    def dedupe_list_of_dict(list_of_dict):
        return [dict(t) for t in {tuple(d.items()) for d in list_of_dict}]

    @staticmethod
    def g_m(m: int):
        if m == 5:
            return "May"
        elif m == 6:
            return "June"
        elif m == 7:
            return "July"
        else:
            return f"{calendar.month_abbr[m]}."

    @staticmethod
    def get_month(month_str: str):
        month_str = month_str.strip()
        if "/" in month_str:
            return "/".join([HoldingsStatementsParser.g_m(int(m)) for m in month_str.split("/")])
        else:
            result = HoldingsStatementsParser.g_m(int(month_str))
            if result == month_str:
                result = HoldingsStatementsParser.g_s(month_str)
            return result

    @staticmethod
    def get_cron_from_to(pattern_field: Field, linked_value_field: Field):
        cron_from = ""
        cron_to = ""
        hlm_stmt = ""
        year = False
        is_span = False
        for chron_level in [cl for cl in "ijkl" if cl in linked_value_field]:
            desc = pattern_field[chron_level] or ""
            if linked_value_field[chron_level]:
                if chron_level == "i" and desc == "(year)":
                    hlm_stmt = linked_value_field[chron_level]
                if desc == "(year)":
                    year = True
                val, *val_rest = linked_value_field[chron_level].split("-")
                is_span = "-" in linked_value_field[chron_level] or is_span
                if desc == "(month)":
                    with contextlib.suppress(Exception):
                        val = HoldingsStatementsParser.get_month(val)
                    with contextlib.suppress(Exception):
                        val_rest = HoldingsStatementsParser.get_month(val_rest[0])
                    if "".join(val_rest):
                        with contextlib.suppress(Exception):
                            val_rest = HoldingsStatementsParser.get_month("".join(val_rest))
                    elif cron_to.strip() and val:
                        val_rest = val
                    if year:
                        spill_year = f"{hlm_stmt}:" if "-" not in hlm_stmt else ""
                        cron_from = f"{cron_from.strip()}:{val} "
                        if cron_to and "".join(val_rest):
                            cron_to = f"{cron_to}:{''.join(val_rest)} "
                        elif not cron_to and "".join(val_rest):
                            cron_to = f"{spill_year}{''.join(val_rest)}"

                else:
                    if "season" in desc:
                        val = HoldingsStatementsParser.get_season(val)
                        if "".join(val_rest):
                            val_rest = HoldingsStatementsParser.get_season("".join(val_rest))
                    cron_from = f"{cron_from} {val}".strip()
                    cron_to = f"{cron_to} {''.join(val_rest)}".strip()
        return (f"{cron_from.strip()}", cron_to.strip(), hlm_stmt, is_span)

    @staticmethod
    def get_from_to(pattern_field: Field, linked_value_field: Field):
        _from = ""
        _to = ""
        is_span = False
        for enum_level in [el for el in "abcdef" if el in linked_value_field]:
            desc = pattern_field[enum_level] or ""
            desc = desc.strip() if "(" not in desc else ""
            if linked_value_field[enum_level]:
                val, *val_rest = linked_value_field[enum_level].split("-")
                is_span = "-" in linked_value_field[enum_level] or is_span
                _from = f"{_from}{(':' if _from else '')}{desc}{val}"
                temp_to = "".join(val_rest)
                if temp_to.strip():
                    _to = f"{_to}{(':' if _to else '')}{desc}{temp_to}"
        return (f"{_from.strip()}", _to.strip(), is_span)



import i18n
from typing import Any
from typing import Dict

from folio_uuid.folio_uuid import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.notes_mapper import NotesMapper
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class CoursesMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        course_map,
        terms_map,
        departments_map,
        library_configuration: LibraryConfiguration,
        task_configuration,
    ):
        self.folio_client: FolioClient = folio_client
        self.user_cache: dict = {}
        self.notes_mapper: NotesMapper = NotesMapper(
            library_configuration,
            self.folio_client,
            course_map,
            FOLIONamespaces.note,
            True,
        )
        self.composite_course_schema = self.get_composite_course_schema()
        self.task_configuration = task_configuration
        super().__init__(
            folio_client,
            self.composite_course_schema,
            course_map,
            None,
            FOLIONamespaces.course,
            library_configuration,
        )
        self.course_map = course_map
        if terms_map:
            self.terms_map = RefDataMapping(
                self.folio_client,
                "/coursereserves/terms",
                "terms",
                terms_map,
                "name",
                "TermsMapping",
            )
        else:
            self.terms_map = None

        if departments_map:
            self.departments_map = RefDataMapping(
                self.folio_client,
                "/coursereserves/departments",
                "departments",
                departments_map,
                "name",
                "DepartmentsMapping",
            )
        else:
            self.departments_map = None

    def store_objects(self, composite_course):
        try:
            self.extradata_writer.write("courselisting", composite_course[0]["courselisting"])
            self.migration_report.add_general_statistics(i18n.t("Stored courselistings"))
            self.extradata_writer.write("course", composite_course[0]["course"])
            self.migration_report.add_general_statistics(i18n.t("Stored courses"))
            if "instructors" in composite_course[0] and any(composite_course[0]["instructors"]):
                for instructor in composite_course[0]["instructors"]:
                    self.extradata_writer.write("instructor", instructor)
                    self.migration_report.add_general_statistics(i18n.t("Stored instructors"))

        except Exception as ee:
            raise TransformationRecordFailedError(
                composite_course[1], "Failed when storing", ee
            ) from ee

    def perform_additional_mappings(self, composite_course):
        try:
            # Assign deterministic ids to every object
            composite_course[0]["course"]["id"] = self.get_uuid(
                composite_course, FOLIONamespaces.course
            )
            composite_course[0]["courselisting"]["id"] = self.get_uuid(
                composite_course, FOLIONamespaces.course_listing
            )
            if "instructors" in composite_course[0] and any(composite_course[0]["instructors"]):
                for idx, instructor in enumerate(composite_course[0]["instructors"]):
                    instructor["id"] = self.get_uuid(
                        composite_course, FOLIONamespaces.instructor, idx
                    )
                    # Link instructor to course listing
                    instructor["courseListingId"] = composite_course[0]["courselisting"]["id"]
                    if self.task_configuration.look_up_instructor:
                        self.populate_instructor_from_users(instructor)
            else:
                self.migration_report.add_general_statistics(i18n.t("Missing Instructors"))

            # Link course to courselisting
            composite_course[0]["course"]["courseListingId"] = composite_course[0][
                "courselisting"
            ]["id"]

        except Exception as ee:
            raise TransformationRecordFailedError(
                composite_course[1], "Failed when creating and linking ids", ee
            ) from ee

    def get_uuid(self, composite_course, object_type: FOLIONamespaces, idx: int = 0):
        return str(
            FolioUUID(
                self.folio_client.okapi_url,
                object_type,
                composite_course[1] if idx == 0 else f"{composite_course[1]}_{idx}",
            )
        )

    def populate_instructor_from_users(self, instructor: dict):
        if instructor["userId"] not in self.user_cache:
            path = "/users"
            query = f'?query=(externalSystemId=="{instructor["userId"]}")'
            if user := next(self.folio_client.folio_get_all(path, "users", query), None):
                self.user_cache[instructor["userId"]] = user
        if user := self.user_cache.get(instructor["userId"], {}):
            instructor["userId"] = user.get("id", "")
            instructor["barcode"] = user.get("barcode", "")
            instructor["patronGroup"] = user.get("patronGroup", "")
        else:
            del instructor["userId"]

    def get_prop(self, legacy_item, folio_prop_name, index_or_id, schema_default_value):
        if folio_prop_name == "courselisting.termId":
            return self.get_mapped_ref_data_value(
                self.terms_map,
                legacy_item,
                folio_prop_name,
                index_or_id,
                False,
            )
        elif folio_prop_name == "course.departmentId":
            return self.get_mapped_ref_data_value(
                self.departments_map,
                legacy_item,
                folio_prop_name,
                index_or_id,
                False,
            )
        elif mapped_value := super().get_prop(
            legacy_item, folio_prop_name, index_or_id, schema_default_value
        ):
            return mapped_value
        else:
            self.migration_report.add("UnmappedProperties", f"{folio_prop_name}")
            return ""

    def get_composite_course_schema(self) -> Dict[str, Any]:
        if self:
            return {
                "properties": {
                    "course": self.folio_client.get_from_github(
                        "folio-org", "mod-courses", "/ramls/course.json"
                    ),
                    "courselisting": self.folio_client.get_from_github(
                        "folio-org", "mod-courses", "/ramls/courselisting.json"
                    ),
                    "instructors": {
                        "type": "array",
                        "items": self.folio_client.get_from_github(
                            "folio-org", "mod-courses", "/ramls/instructor.json"
                        ),
                    },
                }
            }
        else:
            return {
                "properties": {
                    "course": FolioClient.get_latest_from_github(
                        "folio-org", "mod-courses", "/ramls/course.json"
                    ),
                    "courselisting": FolioClient.get_latest_from_github(
                        "folio-org", "mod-courses", "/ramls/courselisting.json"
                    ),
                    "instructors": {
                        "type": "array",
                        "items": FolioClient.get_latest_from_github(
                            "folio-org", "mod-courses", "/ramls/instructor.json"
                        ),
                    },
                }
            }



import json
import logging
import uuid
import i18n
from typing import Any
from typing import Dict
from zoneinfo import ZoneInfo

from dateutil import parser as dateutil_parser
from dateutil import tz
from folio_uuid.folio_uuid import FOLIONamespaces
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class ManualFeeFinesMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        library_configuration: LibraryConfiguration,
        task_configuration,
        feefines_map,
        feefines_owner_map,
        feefines_type_map,
        service_point_map,
        ignore_legacy_identifier: bool = True,
    ):
        self.folio_client: FolioClient = folio_client
        self.composite_feefine_schema = self.get_composite_feefine_schema()
        self.task_configuration = task_configuration
        self.tenant_timezone = self.get_tenant_timezone()

        super().__init__(
            folio_client,
            self.composite_feefine_schema,
            feefines_map,
            None,
            FOLIONamespaces.fees_fines,
            library_configuration,
            ignore_legacy_identifier,
        )

        self.feefines_map = feefines_map
        self.user_cache: dict = {}
        self.item_cache: dict = {}

        if feefines_owner_map:
            self.feefines_owner_map = RefDataMapping(
                self.folio_client,
                "/owners",
                "owners",
                feefines_owner_map,
                "owner",
                "FeeFineOnwerMapping",
            )
        else:
            self.feefines_owner_map = None

        if feefines_type_map:
            self.feefines_type_map = RefDataMapping(
                self.folio_client,
                "/feefines",
                "feefines",
                feefines_type_map,
                "feeFineType",
                "FeeFineTypesMapping",
            )
        else:
            self.feefines_type_map = None

        if service_point_map:
            self.service_point_map = RefDataMapping(
                self.folio_client,
                "/service-points",
                "servicepoints",
                service_point_map,
                "name",
                "FeeFineServicePointTypesMapping",
            )
        else:
            self.service_point_map = None

    def store_objects(self, composite_feefine):
        try:
            self.extradata_writer.write("account", composite_feefine["account"])
            self.migration_report.add_general_statistics(i18n.t("TOTAL Accounts created"))
            self.extradata_writer.write("feefineaction", composite_feefine["feefineaction"])
            self.migration_report.add_general_statistics(i18n.t("TOTAL Feefineactions created"))

        except Exception as ee:
            raise TransformationRecordFailedError(
                composite_feefine, "Failed when storing", ee
            ) from ee

    def get_prop(self, legacy_object, folio_prop_name, index_or_id, schema_default_value):
        if folio_prop_name == "account.ownerId" and self.feefines_owner_map:
            return self.get_mapped_ref_data_value(
                self.feefines_owner_map, legacy_object, index_or_id, folio_prop_name, False
            )

        elif folio_prop_name == "account.feeFineId" and self.feefines_type_map:
            return self.get_mapped_ref_data_value(
                self.feefines_type_map, legacy_object, index_or_id, folio_prop_name, False
            )

        elif folio_prop_name == "feefineaction.createdAt" and self.service_point_map:
            return self.get_mapped_ref_data_value(
                self.service_point_map, legacy_object, index_or_id, folio_prop_name, False
            )

        elif folio_prop_name == "account.amount" or folio_prop_name == "account.remaining":
            return self.parse_sum_as_float(
                index_or_id,
                super().get_prop(
                    legacy_object, folio_prop_name, index_or_id, schema_default_value
                ),
                folio_prop_name,
            )

        elif folio_prop_name == "feefineaction.dateAction":
            return self.parse_date_with_tenant_timezone(
                "feefineaction.dateAction",
                index_or_id,
                super().get_prop(
                    legacy_object, folio_prop_name, index_or_id, schema_default_value
                ),
            )

        elif folio_prop_name == "feefineaction.id":
            return str(uuid.uuid4())

        elif mapped_value := super().get_prop(
            legacy_object, folio_prop_name, index_or_id, schema_default_value
        ):
            return mapped_value

        else:
            return ""

    def get_composite_feefine_schema(self) -> Dict[str, Any]:
        return {
            "properties": {
                "account": FolioClient.get_latest_from_github(
                    "folio-org", "mod-feesfines", "/ramls/accountdata.json"
                ),
                "feefineaction": FolioClient.get_latest_from_github(
                    "folio-org", "mod-feesfines", "/ramls/feefineactiondata.json"
                ),
            }
        }

    def get_tenant_timezone(self):
        config_path = (
            "/configurations/entries?query=(module==ORG%20and%20configName==localeSettings)"
        )
        try:
            tenant_timezone_str = json.loads(
                self.folio_client.folio_get_single_object(config_path)["configs"][0]["value"]
            )["timezone"]
            logging.info("Tenant timezone is: %s", tenant_timezone_str)
            return ZoneInfo(tenant_timezone_str)
        except TypeError as te:
            raise TransformationProcessError(
                "",
                "Failed to fetch Tenant Locale Settings. "
                "Is your library configuration correct?",
            ) from te
        except KeyError as ke:
            raise TransformationProcessError(
                "",
                "Failed to parse Tenant Locale Settings. "
                "Is the Tenant Locale config correctly formatted?",
            ) from ke

    def parse_date_with_tenant_timezone(self, folio_prop_name: str, index_or_id, mapped_value):
        try:
            format_date = dateutil_parser.parse(mapped_value, fuzzy=True)
            if format_date.tzinfo != tz.UTC:
                format_date = format_date.replace(tzinfo=self.tenant_timezone)
            return format_date.isoformat()
        except Exception:
            self.migration_report.add(
                "GeneralStatistics",
                i18n.t("DATA ISSUE Invalid dates"),
            )
            logging.log(
                26,
                "DATA ISSUE\t%s\t%s\t%s",
                index_or_id,
                f"{folio_prop_name} Not a valid date.",
                mapped_value,
            )

    def parse_sum_as_float(self, index_or_id, legacy_sum, folio_prop_name):
        try:
            return float(legacy_sum)
        except Exception as ee:
            self.migration_report.add(
                "GeneralStatistics",
                i18n.t("DATA ISSUE Invalid sum"),
            )
            raise TransformationRecordFailedError(
                index_or_id,
                f"{folio_prop_name} Value must only contain numbers/decimals.",
                legacy_sum,
            ) from ee

    def get_matching_record_from_folio(
        self,
        index_or_id,
        cache: dict,
        path: str,
        match_property: str,
        match_value: str,
        result_type: str,
    ):
        if match_value in cache:
            return cache[match_value]
        else:
            query = f'?query=({match_property}=="{match_value}")'
            if matching_record := next(
                self.folio_client.folio_get_all(path, result_type, query), None
            ):
                cache[match_value] = matching_record
                return matching_record

    def get_folio_user_uuid(self, index_or_id, user_barcode):
        if matching_user := self.get_matching_record_from_folio(
            index_or_id, self.user_cache, "/users", "barcode", user_barcode, "users"
        ):
            return matching_user["id"]
        else:
            self.migration_report.add(
                "GeneralStatistics",
                i18n.t("DATA ISSUE Users not in FOLIO"),
            )
            raise TransformationRecordFailedError(
                index_or_id,
                "No matching user in FOLIO for barcode",
                user_barcode,
            )

    def perform_additional_mapping(self, index_or_id, composite_feefine, legacy_object):
        # Generate account ID
        composite_feefine["account"]["id"] = composite_feefine["id"]
        composite_feefine["feefineaction"]["accountId"] = composite_feefine["id"]

        # Link to FOLIO user
        composite_feefine["account"]["userId"] = self.get_folio_user_uuid(
            index_or_id,
            composite_feefine["account"]["userId"],
        )
        composite_feefine["feefineaction"]["userId"] = composite_feefine["account"]["userId"]

        # Add item data from FOLIO if available
        if item_barcode := composite_feefine["account"].get("itemId"):
            self.enrich_with_folio_item_data(index_or_id, composite_feefine, item_barcode)

        self.add_additional_fields_and_values(composite_feefine, legacy_object)

        return composite_feefine

    def stringify_legacy_object(self, legacy_object):
        legacy_string = (
            "MIGRATION NOTE : This fee/fine was migrated to FOLIO from a previous "
            "library management system. The following is the original data: "
        )
        for key, value in legacy_object.items():
            legacy_string += f"{key.title()}: {value}; "
        return legacy_string.strip().strip(";")

    def enrich_with_folio_item_data(self, index_or_id, feefine, item_barcode):
        if folio_item := self.get_matching_record_from_folio(
            index_or_id,
            self.item_cache,
            "/inventory/items",
            "barcode",
            item_barcode,
            "items",
        ):
            feefine["account"]["itemId"] = folio_item.get("id", "")
            feefine["account"]["title"] = folio_item.get("title", "")
            feefine["account"]["barcode"] = folio_item.get("barcode", "")
            feefine["account"]["callNumber"] = folio_item.get("callNumber", "")
            feefine["account"]["materialType"] = folio_item.get("materialType", {}).get("name")
            feefine["account"]["materialTypeId"] = folio_item.get("materialType", {}).get("id")
            feefine["account"]["location"] = folio_item.get("effectiveLocation", {}).get("name")
        else:
            feefine["account"].pop("itemId")
            self.migration_report.add(
                "GeneralStatistics",
                i18n.t("DATA ISSUE Items not in FOLIO"),
            )
            logging.log(
                26,
                "DATA ISSUE\t%s\t%s\t%s",
                index_or_id,
                "No matching item in FOLIO for barcode",
                item_barcode,
            )

    def add_additional_fields_and_values(self, feefine, legacy_object):
        # Add standard values
        feefine["feefineaction"]["source"] = self.folio_client.username
        feefine["feefineaction"]["notify"] = False
        feefine["feefineaction"]["amountAction"] = feefine["account"]["amount"]
        feefine["feefineaction"]["balance"] = feefine["account"]["remaining"]

        # Set the account status to Open/Closed based on remainign amount
        if feefine["account"]["remaining"] > 0:
            feefine["account"]["status"] = {"name": "Open"}
        else:
            feefine["account"]["status"] = {"name": "Closed"}

        # Add the full legacy item dict to the comment field
        if feefine["feefineaction"].get("comments"):
            feefine["feefineaction"]["comments"] = (
                ("STAFF : " + feefine["feefineaction"]["comments"])
                + " "
                + self.stringify_legacy_object(legacy_object)
            )
        else:
            feefine["feefineaction"]["comments"] = self.stringify_legacy_object(legacy_object)

        # Add name values from reference data mapping
        if self.feefines_owner_map:
            feefine["account"]["feeFineOwner"] = [
                owner["owner"]
                for owner in self.feefines_owner_map.ref_data
                if owner["id"] == feefine["account"]["ownerId"]
            ][0]

        if self.feefines_type_map:
            type_name = [
                type["feeFineType"]
                for type in self.feefines_type_map.ref_data
                if type["id"] == feefine["account"]["feeFineId"]
            ][0]
            feefine["account"]["feeFineType"] = type_name
            feefine["feefineaction"]["typeAction"] = type_name

        return feefine



import csv
import ctypes
import json
import logging
import sys
import time
import i18n
from os.path import isfile
from typing import List
from typing import Optional

from deepdiff import DeepDiff
from folio_uuid.folio_namespaces import FOLIONamespaces
from pydantic.main import BaseModel

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.order_mapper import (
    CompositeOrderMapper,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase

csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))


# Read files and do some work
class OrdersTransformer(MigrationTaskBase):
    class TaskConfiguration(BaseModel):
        name: str
        migration_task_type: str
        files: List[FileDefinition]
        orders_mapping_file_name: str
        organizations_code_map_file_name: str
        acquisition_method_map_file_name: str
        payment_status_map_file_name: Optional[str] = ""
        receipt_status_map_file_name: Optional[str] = ""
        workflow_status_map_file_name: Optional[str] = ""
        location_map_file_name: Optional[str] = ""
        funds_map_file_name: Optional[str] = ""
        funds_expense_class_map_file_name: Optional[str] = ""

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.orders

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        csv.register_dialect("tsv", delimiter="\t")

        super().__init__(library_config, task_config, use_logging)
        self.object_type_name = self.get_object_type().name
        self.task_config = task_config
        self.files = self.list_source_files()
        self.total_records = 0
        self.current_folio_record: dict = {}
        self.orders_map = self.setup_records_map(
            self.folder_structure.mapping_files_folder / self.task_config.orders_mapping_file_name
        )
        self.results_path = self.folder_structure.created_objects_path
        self.failed_files: List[str] = []

        self.folio_keys = []
        self.folio_keys = MappingFileMapperBase.get_mapped_folio_properties_from_map(
            self.orders_map
        )
        self.minted_ids: set = set()

        self.mapper = CompositeOrderMapper(
            self.folio_client,
            self.library_configuration,
            self.orders_map,
            self.load_id_map(self.folder_structure.organizations_id_map_path, True),
            self.load_id_map(self.folder_structure.instance_id_map_path, True),
            self.load_ref_data_mapping_file(
                "acquisitionMethod",
                self.folder_structure.mapping_files_folder
                / self.task_config.acquisition_method_map_file_name,
                self.folio_keys,
            ),
            self.load_ref_data_mapping_file(  # Not required, on POL
                "paymentStatus",
                self.folder_structure.mapping_files_folder
                / self.task_config.payment_status_map_file_name,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(  # Not required, on POL
                "receiptStatus",
                self.folder_structure.mapping_files_folder
                / self.task_config.receipt_status_map_file_name,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(  # Not required
                "workflowStatus",
                self.folder_structure.mapping_files_folder
                / self.task_config.workflow_status_map_file_name,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(
                "locationMap",
                self.folder_structure.mapping_files_folder
                / self.task_config.location_map_file_name,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(  # Required if there was is a fund.
                "fundsMap",
                self.folder_structure.mapping_files_folder / self.task_config.funds_map_file_name,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(  # Todo: The property in the schema has no type
                "fundsExpenseClassMap",
                self.folder_structure.mapping_files_folder
                / self.task_config.funds_expense_class_map_file_name,
                self.folio_keys,
                False,
            ),
        )

    def list_source_files(self):
        files = [
            self.folder_structure.data_folder / self.object_type_name / f.file_name
            for f in self.task_config.files
            if isfile(self.folder_structure.data_folder / self.object_type_name / f.file_name)
        ]
        if not any(files):
            ret_str = ",".join(f.file_name for f in self.task_config.files)
            raise TransformationProcessError(
                f"Files {ret_str} not found in"
                "{self.folder_structure.data_folder} / {self.object_type_name}"
            )
        logging.info("Files to process:")
        for filename in files:
            logging.info("\t%s", filename)
        return files

    def process_single_file(self, filename):
        with open(filename, encoding="utf-8-sig") as records_file, open(
            self.folder_structure.created_objects_path, "w+"
        ) as results_file:
            self.mapper.migration_report.add_general_statistics(
                i18n.t("Number of files processed")
            )
            start = time.time()
            records_processed = 0
            for idx, record in enumerate(self.mapper.get_objects(records_file, filename)):
                records_processed += 1

                try:
                    # Print first legacy record, then first transformed record
                    if idx == 0:
                        logging.info("First legacy record:")
                        logging.info(json.dumps(record, indent=4))

                    folio_rec, legacy_id = self.mapper.do_map(
                        record, f"row {idx}", FOLIONamespaces.orders, True
                    )
                    self.mapper.perform_additional_mapping(legacy_id, folio_rec)

                    self.mapper.migration_report.add_general_statistics(
                        i18n.t("TOTAL Purchase Order Lines created")
                    )
                    self.mapper.report_folio_mapping(folio_rec, self.mapper.composite_order_schema)
                    self.mapper.notes_mapper.map_notes(
                        record,
                        legacy_id,
                        folio_rec["id"],
                        FOLIONamespaces.orders,
                    )

                    self.merge_into_orders_with_embedded_pols(folio_rec, results_file)

                except TransformationProcessError as process_error:
                    self.mapper.handle_transformation_process_error(idx, process_error)
                except TransformationRecordFailedError as error:
                    self.mapper.handle_transformation_record_failed_error(idx, error)
                except Exception as excepion:
                    self.mapper.handle_generic_exception(idx, excepion)

                # TODO Rewrite to base % value on number of rows in file
                if idx > 1 and idx % 50 == 0:
                    elapsed = idx / (time.time() - start)
                    elapsed_formatted = "{0:.4g}".format(elapsed)
                    logging.info(  # pylint: disable=logging-fstring-interpolation
                        f"{idx:,} records processed. Recs/sec: {elapsed_formatted} "
                    )

            self.total_records = records_processed

            logging.info(  # pylint: disable=logging-fstring-interpolation
                f"Done processing {filename} containing {self.total_records:,} records. "
                f"Total records processed: {self.total_records:,}"
            )
            logging.info("Storing last record to disk")
            Helper.write_to_file(results_file, self.current_folio_record)
            self.mapper.migration_report.add_general_statistics(
                i18n.t("TOTAL Purchase Orders created")
            )

    def do_work(self):
        logging.info("Getting started!")
        for file in self.files:
            logging.info("Processing %s", file)
            try:
                print(file)
                self.process_single_file(file)
            except Exception as ee:
                error_str = (
                    f"Processing of {file} failed:\n{ee}."
                    "Check source files for empty lines or missing reference data"
                )
                logging.exception(error_str)
                self.mapper.migration_report.add("FailedFiles", f"{file} - {ee}")
                sys.exit()

    def wrap_up(self):
        logging.info("Done. Wrapping up...")
        with open(self.folder_structure.migration_reports_file, "w") as migration_report_file:
            logging.info(
                "Writing migration- and mapping report to %s",
                self.folder_structure.migration_reports_file,
            )
            self.mapper.migration_report.write_migration_report(
                i18n.t("Pruchase Orders and Purchase Order Lines Transformation Report"),
                migration_report_file,
                self.start_datetime,
            )

            Helper.print_mapping_report(
                migration_report_file,
                self.total_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )
        logging.info("All done!")

    def merge_into_orders_with_embedded_pols(self, folio_rec, results_file):
        # Handle merging and storage
        if not self.current_folio_record:
            self.current_folio_record = folio_rec
        if folio_rec["id"] != self.current_folio_record["id"]:
            # Writes record to file
            Helper.write_to_file(results_file, self.current_folio_record)
            self.mapper.migration_report.add_general_statistics(
                i18n.t("TOTAL Purchase Orders created")
            )
            self.current_folio_record = folio_rec

        else:
            # Merge if possible
            diff = DeepDiff(self.current_folio_record, folio_rec)
            if "compositePoLines" in diff.affected_root_keys:
                self.current_folio_record.get("compositePoLines", []).extend(
                    folio_rec.get("compositePoLines", [])
                )
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Rows merged to create Purchase Orders")
                )
            for key in diff.affected_paths:
                self.mapper.migration_report.add("DiffsBetweenOrders", key)
